@@ -28,7 +28,7 @@ static void	  gail_label_real_initialize	   (AtkObject 	      *obj,
                                                     gpointer	      data);
 static void	  gail_label_real_notify_gtk	   (GObject	      *obj,
                                                     GParamSpec	      *pspec);
-static gint       gail_label_map_gtk               (GtkWidget         *widget,
+static void       gail_label_map_gtk               (GtkWidget         *widget,
                                                     gpointer          data);
 static void       gail_label_init_text_util        (GailLabel         *gail_label,
                                                     GtkWidget         *widget);
@@ -163,7 +163,6 @@ static void
 gail_label_real_initialize (AtkObject *obj,
                             gpointer  data)
 {
-  GtkLabel  *label;
   GtkWidget  *widget;
   GailLabel *gail_label;
 
@@ -208,7 +207,7 @@ gail_label_real_initialize (AtkObject *obj,
     obj->role = ATK_ROLE_LABEL;
 }
 
-static gint
+static void
 gail_label_map_gtk (GtkWidget *widget,
                     gpointer data)
 {
@@ -276,6 +275,27 @@ notify_name_change (AtkObject *atk_obj)
   if (gail_label->textutil == NULL)
     return;
 
+  /*
+   * Check whether the label has actually changed before emitting
+   * notification.
+   */
+  if (gail_label->textutil->buffer) 
+    {
+      GtkTextIter start, end;
+      const char *new_label;
+      char *old_label;
+      int same;   
+
+      gtk_text_buffer_get_start_iter (gail_label->textutil->buffer, &start);
+      gtk_text_buffer_get_end_iter (gail_label->textutil->buffer, &end);
+      old_label = gtk_text_buffer_get_text (gail_label->textutil->buffer, &start, &end, FALSE);
+      new_label = gtk_label_get_text (label);
+      same = strcmp (new_label, old_label);
+      g_free (old_label);
+      if (same == 0)
+        return;
+    }
+   
   /* Create a delete text and an insert text signal */
  
   g_signal_emit_by_name (gail_obj, "text_changed::delete", 0, 
@@ -330,6 +350,7 @@ gail_label_real_notify_gtk (GObject           *obj,
       if (!gail_label->has_top_level)
         {
           temp_obj = atk_obj;
+          top_level = NULL;
           while (temp_obj)
             {
               top_level = temp_obj;
@@ -467,14 +488,34 @@ gail_label_ref_relation_set (AtkObject *obj)
           AtkObject *accessible_array[1];
           AtkRelation* relation;
 
-          /*
-           * Handle the case where a GnomeIconEntry is specified as the 
-           * mnemonic widget. use the button which is a grandchild of the
-           * GnomeIconEntry as the mnemonic widget. See bug #133967.
-           */
           if (!GTK_WIDGET_CAN_FOCUS (mnemonic_widget))
             {
-              if (GTK_IS_BOX (mnemonic_widget))
+            /*
+             * Handle the case where a GtkFileChooserButton is specified as the 
+             * mnemonic widget. use the combobox which is a child of the
+             * GtkFileChooserButton as the mnemonic widget. See bug #359843.
+             */
+             if (GTK_IS_BOX (mnemonic_widget))
+               {
+                  GList *list, *tmpl;
+
+                  list = gtk_container_get_children (GTK_CONTAINER (mnemonic_widget));
+                  if (g_list_length (list) == 2)
+                    {
+                      tmpl = g_list_last (list);
+                      if (GTK_IS_COMBO_BOX(tmpl->data))
+                        {
+                          mnemonic_widget = GTK_WIDGET(tmpl->data);
+                        }
+                    }
+                  g_list_free (list);
+                }
+            /*
+             * Handle the case where a GnomeIconEntry is specified as the 
+             * mnemonic widget. use the button which is a grandchild of the
+             * GnomeIconEntry as the mnemonic widget. See bug #133967.
+             */
+              else if (GTK_IS_BOX (mnemonic_widget))
                 {
                   GList *list;
 
@@ -498,7 +539,7 @@ gail_label_ref_relation_set (AtkObject *obj)
                           list = gtk_container_get_children (GTK_CONTAINER (temp_widget));
                           if (GTK_IS_COMBO (list->data))
                             {
-                              mnemonic_widget = GTK_COMBO (list->data);
+                              mnemonic_widget = GTK_WIDGET (list->data);
                             }
                         }
                     }
@@ -596,6 +637,8 @@ gail_label_get_text (AtkText *text,
     return NULL;
   else
   {
+    if (GAIL_LABEL (text)->textutil == NULL) 
+      gail_label_init_text_util (GAIL_LABEL (text), widget);
     return gail_text_util_get_substring (GAIL_LABEL(text)->textutil, 
                                          start_pos, end_pos);
   }
